-- ----------------------------------------------------------------------------
-- Copyright (C) 2013 - 2013 HERE Global B.V., including its affiliated companies.
--
-- These coded instructions, statements, and computer programs contain
-- unpublished proprietary information of HERE Global B.V., and
-- are copy protected by law. They may not be disclosed to third parties
-- or copied or duplicated in any form, in whole or in part, without the
-- specific, prior written permission of HERE Global B.V.
-- ----------------------------------------------------------------------------
--      Author: Dirk Huppermann, Bill King
-- ----------------------------------------------------------------------------
--      Output format for Nuance Engine
--      0.4.0.XXX


_package_settings["phonemeset"] = "NT-SAMPA"

marc_to_nuance = {
    ["ENG"] = "eng",  -- UK English
    ["010"] = "enu",  -- US English
    ["FRE"] = "frf",
    ["GER"] = "ged",
    ["ITA"] = "iti",
    ["SPA"] = "spe"
}

phoneme_lang_to_marc = {
    ["UKE"] = "ENG",   -- UK English
    ["ENG"] = "010",   -- US English
    ["FRE"] = "FRE",
    ["GER"] = "GER",
    ["ITA"] = "ITA",
    ["SPA"] = "SPA"
}

function take_phoneme_if_exists( name, phoneme, language )
    result = ""
    -- remove "or ( language == "ENG" and marc_code == "ENG" )" once issue NAV-4513 is fixed
    if use_phoneticdata
        and phoneme ~= nil
        and phoneme ~= ""
        and ( ( phoneme_lang_to_marc[language] == marc_code ) or ( language == "ENG" and marc_code == "ENG" ) )
    then
        -- for NT-SAMPA phonemes
        result = result .. " <ESC>\\tn=address\\ <ESC>\\toi=nts:\""..name.."\"\\ "..phoneme.." <ESC>\\toi=orth\\ <ESC>\\tn=normal\\ "
    else
        result = ' <ESC>\\tn=address\\ '..name..' <ESC>\\tn=normal\\ '
    end
--    if (language ~= nil and string.len(language) == 3 ) then
--        local dest_lang = marc_to_nuance[phoneme_lang_to_marc(language)] or ''
--        local orig_lang = marc_to_nuance[marc_code] or ''
--        result = "<ESC>\\lang="..dest_lang.."\\ " .. result .. "<ESC>\\lang="..orig_lang.."\\ "
--    end
    return result
end

function esc(x)
  return (x:gsub('%%', '%%%%')
           :gsub('%^', '%%%^')
           :gsub('%$', '%%%$')
           :gsub('%(', '%%%(')
           :gsub('%)', '%%%)')
           :gsub('%.', '%%%.')
           :gsub('%[', '%%%[')
           :gsub('%]', '%%%]')
           :gsub('%*', '%%%*')
           :gsub('%+', '%%%+')
           :gsub('%-', '%%%-')
           :gsub('%?', '%%%?'))
end

function set_result( id )

    if nonTTS and double_command and id == command_id_1 and commands_double[ id ] ~= nil and commands_double[ id ] ~= "" then
        result_sentence = commands_double[ id ]
    elseif guidance_mode == "pedestrian" and commands_ped[ id ] ~= nil and commands_ped[ id ] ~= "" then
        result_sentence = commands_ped[ id ]
    else
        result_sentence = commands_common[ id ]
    end

    if result_sentence ~= nil and result_sentence ~= "" and not nonTTS then

        tts_street_1 = build_streetname(tts_street_1, result_sentence, take_phoneme_if_exists)
        tts_street_2 = build_streetname(tts_street_2, result_sentence, take_phoneme_if_exists)

        tts_signpost_1 = build_signpost(tts_signpost_1, result_sentence, take_phoneme_if_exists)
        tts_signpost_2 = build_signpost(tts_signpost_2, result_sentence, take_phoneme_if_exists)

        tts_exit_number = build_signpost(tts_exit_number, result_sentence, take_phoneme_if_exists)

        natural_guidance_terms = provide_natual_guidance( {maneuver, nextManeuver}, take_phoneme_if_exists)

        result_sentence, trash = string.gsub ( result_sentence, "!FREQUENCY!", frequency )

        if string.find( result_sentence, "!SET_AUDIO!" ) then
            audio_start = '<ESC>\\audio="'..audio_files_path_absolute..package.config:sub(1, 1)
            audio_end = "\"\\"
            result_sentence, trash = string.gsub ( result_sentence, "!SET_AUDIO!(.+)!SET_AUDIO!", audio_start.."%1"..audio_end )
        end

        result_sentence, trash = string.gsub ( result_sentence, "!PED_TURN_NO!", ped_turn_no )
        result_sentence, trash = string.gsub ( result_sentence, "!DIST!", distance_1 )
        result_sentence, trash = string.gsub ( result_sentence, "!UNIT!", usedUnit_1 )
        result_sentence, trash = string.gsub ( result_sentence, "!STREET!", esc(tts_street_1) )
        result_sentence, trash = string.gsub ( result_sentence, "!STREET_2!", esc(tts_street_2) )
        result_sentence, trash = string.gsub ( result_sentence, "!SIGNPOST!", esc(tts_signpost_1) )
        result_sentence, trash = string.gsub ( result_sentence, "!SIGNPOST_2!", esc(tts_signpost_2) )
        result_sentence, trash = string.gsub ( result_sentence, "!EXIT_NO_ROUNDABOUT!", exit_no_roundabout )
        result_sentence, trash = string.gsub ( result_sentence, "!ORIENTATION!", head_orientation )
        result_sentence, trash = string.gsub ( result_sentence, "!NG_COMMAND_1!", esc(natural_guidance_terms[1]) )
        result_sentence, trash = string.gsub ( result_sentence, "!NG_COMMAND_2!", esc(natural_guidance_terms[2]) )
        result_sentence, trash = string.gsub ( result_sentence, "!EXIT_NUMBER!", esc(tts_exit_number) )
        result_sentence, trash = string.gsub ( result_sentence, "!BREAK!", "<ESC>\\pause=100\\" )

        while string.find( result_sentence, "!BREAK:" ) do
            pauseTime, trash = string.gsub(result_sentence, ".+!BREAK:(%d+).+", "%1")
            result_sentence, trash = string.gsub ( result_sentence, "!BREAK:"..pauseTime.."!", "<ESC>\\pause="..pauseTime.."\\", 1)
        end

        while true do
            result_sentence, b_use = string.gsub ( result_sentence, "  ", " " )
            if b_use == 0 then
                break
            end
        end

        if ( string.byte(result_sentence,-1) == 32 ) then
            result_sentence = string.sub(result_sentence,1,-2)
        end

        result_sentence, trash = string.gsub( result_sentence, " ,", "," )

        result_sentence = languageSpecificMods( result_sentence )
        
        -- insert comma between two consecutive phonemes( e.g. between the exit_number and exit_name )
        result_sentence, trash = string.gsub(result_sentence, " <ESC>\\tn=normal\\ <ESC>\\tn=address\\", ",")

        result_sentence, trash = string.gsub( result_sentence, "<ESC>", "")
        
        -- remove comma from the end of the sentence (useful when natural guidance elements are added to the prompt)
        result_sentence, trash = string.gsub( result_sentence, ",$", "")
        
        -- trim front and back
        result_sentence, trash = string.gsub(result_sentence, "^(%s+)(.+)(%s*)$", "%2")

    elseif nonTTS then

        result_sentence, trash = string.gsub ( result_sentence, "!DIST!", distance_1 )
        result_sentence, trash = string.gsub ( result_sentence, "!UNIT!", usedUnit_1 )
        result_sentence, trash = string.gsub ( result_sentence, "!EXIT_NO_ROUNDABOUT!", exit_no_roundabout )

    end

    return result_sentence

end

function set_result_tts( )

    if double_command then
        if sentence_1 ~= "" and sentence_1 ~= " " and sentence_1 ~= nil then
            result_sentence = sentence_1.." "..sentence_2
        else
            result_sentence = ""
        end
    else
        if sentence_1 ~= "" and sentence_1 ~= " " and sentence_1 ~= nil then
            result_sentence = sentence_1
        else
            result_sentence = ""
        end
    end

    result_sentence, b_use = string.gsub ( result_sentence, "  ", "" )

    if ( string.byte(result_sentence,-1) == 32 ) then
        result_sentence = string.sub(result_sentence,1,-2)
    end

    result_sentence = string.upper( string.sub(result_sentence, 1, 1 ) )..string.sub(result_sentence, 2 )

    result_sentence, count = string.gsub( result_sentence, '(audio=\\"[%a%d]+.wav"\\)'..'(%.)', "%1" )
    result_sentence, count = string.gsub( result_sentence, ',%p', "." ) -- Shouldn't have to handle this, have a pre-check.
    table.insert( result_list, result_sentence)

end

function set_result_wave( )

    for k,v in pairs(sentence_1) do
        table.insert( result_list, v )
    end

    if double_command then
        for k,v in pairs(sentence_2) do
            table.insert( result_list, v )
        end
    end

end
